from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.shortcuts import reverse
from django.utils.translation import gettext_lazy as _

from djmoney.models.fields import CurrencyField, MoneyField
from payments import PurchasedItem
from payments.models import BasePayment

from aleksis.core.mixins import ExtensibleModel, PureDjangoModel
from aleksis.core.models import Person

from .base import Client
from ..tables import PurchasedItemsTable, TotalsTable


class InvoiceGroup(ExtensibleModel):
    name = models.CharField(verbose_name=_("Invoice group name"), max_length=255)
    client = models.ForeignKey(
        Client, verbose_name=_("Linked client"), related_name="invoice_groups", on_delete=models.SET_NULL, null=True
    )

    template_name = models.CharField(verbose_name=_("Template to render invoices with as PDF"), blank=True, max_length=255)

    def __str__(self) -> str:
        return self.name


    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["client", "name"], name="group_uniq_per_client")
        ]


class Invoice(BasePayment, PureDjangoModel):
    VARIANT_DISPLAY = {
        "paypal": (_("PayPal"), "logos:paypal"),
        "sofort": (_("Klarna / Sofort"), "simple-icons:klarna"),
        "pledge": (_("Payment pledge / manual payment"), "mdi:hand-coin"),
        "sdd": (_("SEPA Direct Debit"), "mdi:bank-transfer"),
    }

    group = models.ForeignKey(
        InvoiceGroup, verbose_name=_("Invoice group"), related_name="invoices", on_delete=models.SET_NULL, null=True
    )

    number = models.CharField(verbose_name=_("Invoice number"), max_length=255)

    for_content_type = models.ForeignKey(ContentType, on_delete=models.SET_NULL, null=True)
    for_object_id = models.PositiveIntegerField()
    for_object = GenericForeignKey("for_content_type", "for_object_id")

    @classmethod
    def get_variant_choices(cls):
        choices = []
        for variant in settings.PAYMENT_VARIANTS.keys():
            choices.append((variant, cls.VARIANT_DISPLAY[variant][0]))
        return choices

    def get_variant_name(self):
        return self.__class__.VARIANT_DISPLAY[self.variant][0]

    def get_variant_icon(self):
        return self.__class__.VARIANT_DISPLAY[self.variant][1]

    def get_purchased_items(self):
        return self.for_object.get_purchased_items()

    def get_person(self):
        if hasattr(self.for_object, "person"):
            return self.for_object.person
        elif hasattr(self.for_object, "get_person"):
            return self.for_object.get_person()

        return None

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["number", "group"], name="number_uniq_per_group")
        ]

    @property
    def purchased_items_table(self):
        items = [i._asdict() for i in self.get_purchased_items()]
        return PurchasedItemsTable(items)

    @property
    def totals_table(self):
        tax_amounts = {}
        for item in self.get_purchased_items():
            tax_amounts.setdefault(item.tax_rate, 0)
            tax_amounts[item.tax_rate] += item.price / (item.tax_rate + 100) * item.tax_rate

        values = []
        for tax_rate, total in tax_amounts.items():
            values.append({
                "name": _("Included VAT {} %").format(tax_rate),
                "value": total,
                "currency": self.currency,
            })

        values.append({
            "name": _("Gross total"),
            "value": self.total,
            "currency": self.currency,
        })

        return TotalsTable(values)

    def get_success_url(self):
        return reverse("invoice_by_token", kwargs={"slug": self.token})

    def get_failure_url(self):
        return reverse("invoice_by_token", kwargs={"slug": self.token})


class ManualInvoice(ExtensibleModel):
    invoice_group = models.ForeignKey(InvoiceGroup, on_delete=models.CASCADE, verbose_name=_("Invoice group"))
    person = models.ForeignKey(Person, on_delete=models.SET_NULL, verbose_name=_("Invoice recipient (person)"))

    item_description = models.CharField(max_length=255, verbose_name=_("Purchased item"))
    price = models.DecimalField(verbose_name=_("Item gross price"), max_digits=9, decimal_places=2, default="0.0")
    currency = models.CharField(max_length=10, verbose_name=_("Currency"))
    tax_rate = models.DecimalField(verbose_name=_("Tax rate"), max_digits=4, decimal_places=1, default="0.0")

    variant = models.CharField(verbose_name=_("Payment variant"), max_length=255, blank=True)

    def get_purchased_items(self):
        yield PurchasedItem(name=self.item_description, quantity=1, price=self.price, currency=self.currency, sku="", tax_rate=self.tax_rate)

    def get_invoice(self):
        invoice, __ = Invoice.objects.get_or_create(for_content_type=ContentType.objects.get_for_model(self), for_object_id=self.pk, defaults={
            "group": self.invoice_group,
            "variant": self.variant,
            "number": f"FIXME",
            "currency": self.currency,
            "total": self.price,
            "tax": self.total / (self.tax_rate + 100) * self.tax_rate,
            "description": self.item_description,
            "billing_first_name": self.person.first_name,
            "billing_last_name": self.person.last_name,
            "billing_address_1": f"{self.person.street} {self.person.housenumber}",
            "billing_city": self.person.place,
            "billing_postcode": self.person.postal_code,
            "billing_email": self.person.email,
        })

        return invoice
